#include "./qrcodes.h"

void print_usage(char *name) {
  printf("Usage:\n");
  printf("  %s options\n", name);
  printf("    -h            this message\n");
  printf("    -d filename   decodes the QR code contained in filename\n");
  printf("                  (pbm file)\n");
  printf("    -i filename   gives infromation about QR code contained in filename\n");
  printf("                  (pbm file)\n");
  printf("    -m nb_pixel   size of margin, in pixels, around the QR codes (default: 12)\n");
  printf("    -c nb_pixel   size of modules (cells), in pixels, of the QR codes (default: 3)\n");
//  printf("    -v            increases verbosity\n");
  printf("\n");
  printf("    -T            perform some tests (debug only)\n");
  printf("                  ignore all other flags");
  printf("\n");
}



/* nombre d'octets de données pour chacun des niveaux de correction d'erreurs */
int NB_DATA_BYTES[4] = { 16, 19, 9, 13};

//int verbosity = 0;

int main(int argc, char *argv[]){

  /* initialisation des table de log et puissances */
  init_tables();

  int opt;
  int margin_size = 12;
  int module_size = 3;
  char filename[80];
  int decode_filename = 0;
  int info_filename = 0;

  if (argc == 1) {
    print_usage(argv[0]);
    exit(-1);
  }

  while ((opt = getopt(argc, argv, "hTd:i:m:c:")) != -1) {
    switch (opt) {
      case 'h':
        print_usage(argv[0]);
        exit(0);
        break;
//      case 'v':
//        verbosity++;
//        break;
      case 'T':
        do_tests();
        exit(0);
        break;
      case 'm':
        margin_size = atoi(optarg);
        break;
      case 'c':
        module_size = atoi(optarg);
        break;
      case 'd':
        strncpy(filename, optarg, 80);
        decode_filename = 1;
        info_filename = 0;
        break;
      case 'i':
        strncpy(filename, optarg, 80);
        decode_filename = 0;
        info_filename = 1;
        break;
      default: /* '?' */
        printf("%s -h   to get a help message\n", argv[0]);
        exit(-1);
    }
  }



  if (decode_filename != 0) {
    decode_file(filename, margin_size, module_size);
    return(0);
  }
  if (info_filename != 0) {
    info_file(filename, margin_size, module_size);
    return(0);
  }

  return 0;
}

/* vim: set tabstop=2 shiftwidth=2 : */
