#include "./qrcodes.h"

/***************************************************************
 * création des fichiers pour les differents masques possibles *
 ***************************************************************/
#include <pgm.h>
void create_mask_images(void) {
  char FIXED[21][21] = {
    {1,1,1,1,1,1,1,0,3,2,2,2,2,0,1,1,1,1,1,1,1},
    {1,0,0,0,0,0,1,0,3,2,2,2,2,0,1,0,0,0,0,0,1},
    {1,0,1,1,1,0,1,0,3,2,2,2,2,0,1,0,1,1,1,0,1},
    {1,0,1,1,1,0,1,0,3,2,2,2,2,0,1,0,1,1,1,0,1},
    {1,0,1,1,1,0,1,0,3,2,2,2,2,0,1,0,1,1,1,0,1},
    {1,0,0,0,0,0,1,0,3,2,2,2,2,0,1,0,0,0,0,0,1},
    {1,1,1,1,1,1,1,0,1,0,1,0,1,0,1,1,1,1,1,1,1},
    {0,0,0,0,0,0,0,0,3,2,2,2,2,0,0,0,0,0,0,0,0},
    {3,3,3,3,3,3,1,3,3,2,2,2,2,3,3,3,3,3,3,3,3},
    {2,2,2,2,2,2,0,2,2,2,2,2,2,2,2,2,2,2,2,2,2},
    {2,2,2,2,2,2,1,2,2,2,2,2,2,2,2,2,2,2,2,2,2},
    {2,2,2,2,2,2,0,2,2,2,2,2,2,2,2,2,2,2,2,2,2},
    {2,2,2,2,2,2,1,2,2,2,2,2,2,2,2,2,2,2,2,2,2},
    {0,0,0,0,0,0,0,0,1,2,2,2,2,2,2,2,2,2,2,2,2},
    {1,1,1,1,1,1,1,0,3,2,2,2,2,2,2,2,2,2,2,2,2},
    {1,0,0,0,0,0,1,0,3,2,2,2,2,2,2,2,2,2,2,2,2},
    {1,0,1,1,1,0,1,0,3,2,2,2,2,2,2,2,2,2,2,2,2},
    {1,0,1,1,1,0,1,0,3,2,2,2,2,2,2,2,2,2,2,2,2},
    {1,0,1,1,1,0,1,0,3,2,2,2,2,2,2,2,2,2,2,2,2},
    {1,0,0,0,0,0,1,0,3,2,2,2,2,2,2,2,2,2,2,2,2},
    {1,1,1,1,1,1,1,0,3,2,2,2,2,2,2,2,2,2,2,2,2}
  };

  int i,j;
  int x,y;
  int module_size = 3;
  bit **qr = alloc_qrcode(21);
  int taille = (8+SIZE_V1)*module_size;
  gray **T = malloc(taille*sizeof(gray*));
  for (i=0; i<taille; i++)
    T[i] = malloc(taille*sizeof(gray));
  char image[256];

  for (int m=0; m<8; m++) {
    sprintf(image, "masque-%d.pgm",m);
    for (i=0; i<21; i++)
      for (j=0; j<21; j++)
        qr[i][j] = 0;
    data_mask(qr, m);

    for (i=0; i<taille; i++) {
      for (j=0; j<taille; j++) {
        if (i<4*module_size || i>=(SIZE_V1+4)*module_size
            || j<4*module_size || j>=(SIZE_V1+4)*module_size) {
          T[i][j] = 255;
        } else {
          x = (i-4*module_size)/module_size;
          y = (j-4*module_size)/module_size;
          if (qr[x][y] == 0) T[i][j] = 255;
          if (qr[x][y] == 1) T[i][j] = 0;
          if (FIXED[x][y] == 0) T[i][j] = 177;
          if (FIXED[x][y] == 1) T[i][j] = 87;
          if (FIXED[x][y] == 3) T[i][j] = 127;
        }
      }
    }
    FILE *img = fopen(image, "w");
    if (img == NULL) {
      fprintf(stderr, "Cannot open file \"%s\" for writing. Aborting...\n", image);
      exit(-1);
    }
    pgm_writepgm(img, T, taille, taille, 255, 1);
    fclose(img);
  }
  for (i=0; i<taille; i++) free(T[i]);
  free(T);
}

/**********************************************************
 * divers tests pour les fonction (utilisés pour les TPs) *
 **********************************************************/

void test_check_fixed(char *filename) {
  QRCode qr = alloc_qrcode(SIZE_V1);
  pbm_to_qrcode_version1(filename, 12, 3, qr);
  int nb_erreurs = check_fixed_bits_version1(qr);
  printf("fichier          : \"%s\"\n", filename);
  printf("nombre d'erreurs : %d\n", nb_erreurs);
  printf("\n");
}

void test_raw_format(char *filename) {
  QRCode qr = alloc_qrcode(SIZE_V1);
  pbm_to_qrcode_version1(filename, 12, 3, qr);
  bit format[2][15];
  get_raw_format_bits(qr, format);
  printf("fichier    : \"%s\"\n", filename);
  printf("format - 1 : "); print_bits_array(format[0], 15);
  printf("format - 2 : "); print_bits_array(format[1], 15);
  printf("\n");
}

void test_format(char *filename) {
  QRCode qr = alloc_qrcode(SIZE_V1);
  pbm_to_qrcode_version1(filename, 12, 3, qr);
  bit format[2][15];
  get_format_bits(qr, format);
  printf("fichier    : \"%s\"\n", filename);
  printf("format - 1 : "); print_bits_array(format[0], 15);
  printf("format - 2 : "); print_bits_array(format[1], 15);
  int d = correct_format(format[0]);
  printf("%d erreurs, ",d);
  printf("format - 1 : "); print_bits_array(format[0], 15);
  d = correct_format(format[1]);
  printf("%d erreurs, ",d);
  printf("format - 2 : "); print_bits_array(format[1], 15);
  printf("\n");
}

void test_level_mask(char *filename) {
  QRCode qr = alloc_qrcode(SIZE_V1);
  pbm_to_qrcode_version1(filename, 12, 3, qr);
  int l, m;
  get_format(qr, &l, &m);
  printf("fichier      : \"%s\"\n", filename);
  printf("level / mask : %d / %d\n", l, m);
  printf("\n");
}

void test_masking(char *filename, char *new) {
  QRCode qr = alloc_qrcode(SIZE_V1);
  pbm_to_qrcode_version1(filename, 12, 3, qr);
  int l, m;
  get_format(qr, &l, &m);
  data_mask(qr, m);
  qrcode_version1_to_pbm(new, 3, qr);
}

void test_data(char *filename) {
  QRCode qr = alloc_qrcode(SIZE_V1);
  pbm_to_qrcode_version1(filename, 12, 3, qr);
  int l, m;
  get_format(qr, &l, &m);
  data_mask(qr, m);
  byte data[26];
  get_data_bytes(qr, data);
  printf("fichier : \"%s\"\n", filename);
  printf("données : "); for (int i=0; i<26; i++) printf("%02x ", data[i]); printf("\n");
}

void test_data_syndromes(char *filename) {
  QRCode qr = alloc_qrcode(SIZE_V1);
  pbm_to_qrcode_version1(filename, 12, 3, qr);
  int l, m;
  get_format(qr, &l, &m);
  data_mask(qr, m);
  byte data[26];
  get_data_bytes(qr, data);
  printf("fichier : \"%s\"\n", filename);
  int nb_redundant_bytes = SIZE_V1 - NB_DATA_BYTES[l];
  printf("nombre d'octets de redondances = %d\n", nb_redundant_bytes);
  byte *S = malloc(nb_redundant_bytes);
  data_syndrome(data, nb_redundant_bytes, S);
  printf("syndromes :\n");
  for (int i=0; i<nb_redundant_bytes; i++) printf("%02x, ", S[i]); printf("\n");
}

void test_mode_length(char *filename) {
  QRCode qr = alloc_qrcode(SIZE_V1);
  pbm_to_qrcode_version1(filename, 12, 3, qr);
  int l, m;
  get_format(qr, &l, &m);
  data_mask(qr, m);
  byte data[26];
  get_data_bytes(qr, data);
  int mode, length;
  mode = get_mode(data, 0);
  length = get_length(data, 0, mode);
  printf("fichier       : \"%s\"\n", filename);
  printf("mode / length : %d / %d\n", mode, length);
}

void test_solve_linear_system(void) {
  int taille = 3;
  byte **S = malloc((taille+1)*sizeof(byte*));
  int i,j;
  for (i=0; i<taille+1; i++)
    S[i] = malloc((taille+1)*sizeof(byte));
  byte *result = malloc(sizeof(byte)*(taille+1));

  /*
  byte tmp[3][4] = {
    { 0x20, 0x31, 0x07, 0x36},
    { 0x12, 0xfa, 0xff, 0x88},
    { 0xa5, 0x00, 0x76, 0xa8}
  };
  */
  /*
  byte tmp[3][4] = {
    { 0x00, 0x00, 0x01, 0xaa},
    { 0x00, 0x01, 0x00, 0xbb},
    { 0x01, 0x00, 0x00, 0xcc}
  };
  */
  byte tmp[4][4] = {
    { 0x01, 0x57, 0x4a, 0x20},
    { 0x20, 0x31, 0x07, 0x36},
    { 0x12, 0xfa, 0xff, 0x88},
    { 0xa5, 0x00, 0x76, 0xa8},
  };

  for (i=0; i<taille+1; i++)
    for (j=0; j<taille+1; j++)
      S[i][j] = tmp[i][j];

  int r = solve_linear_system(taille+1, taille+1, S, result);

  printf("r = %d\n", r);
  for (i=0; i<taille; i++)
    printf("X[%d] = 0x%02x\n", i, result[i]);
}

/* définis dans correction.c */
byte EXP[256];
byte LOG[256];

void show_tables(void) {
  init_tables();
  int i;
  printf("EXP:");
  for (i=0; i<256; i++) {
    if (i%16 == 0) printf("\n");
    printf("  %02x", EXP[i]);
  }
  printf("\n\nLOG:");
  for (i=0; i<256; i++) {
    if (i%16 == 0) printf("\n");
    printf("  %02x", LOG[i]);
  }
  printf("\n\n");
}

void test_tables(void) {
  init_tables();
  for (int i=0; i<256; i++) {
    printf("exp(log(%02x)) = %02x\n",EXP[LOG[i]],i);
    printf("log(exp(%02x)) = %02x\n",LOG[EXP[i]],i);
  }
}


void test_correct_data(char *filename) {
  printf("fichier %s :\n", filename);
  QRCode qr = alloc_qrcode(SIZE_V1);
  pbm_to_qrcode_version1(filename, 12, 3, qr);
  int level, mask;
  get_format(qr, &level, &mask);
  data_mask(qr, mask);
  byte data[26];
  get_data_bytes(qr, data);

  correct_data(data, level);
    /* on récupère combien d'octets sont bien des octets de données (par
     * opposition aux octets de correction) */
  int nb_data_bytes = NB_DATA_BYTES[level];
    /* on décode... */
    char *mes = malloc(42); /* il y a au plus 41 caractères dans un
                               QR code version 1... */
    int mode;
    int length;

    int offset = 0;
    int nb_char = 0;
    int end = 0;
    while(0 == end && offset < 8*nb_data_bytes-4) {
      mode = get_mode(data,offset);
      switch (mode) {
        case (BYTE_MODE):
          length = get_length(data,offset,mode);
          offset += decode_bytes(data, 0, length, mes+nb_char) + 12;
          nb_char += length;
          break;
        case (NUMERIC_MODE):
          length = get_length(data,offset,mode);
          offset += decode_numeric(data, 0, length, mes+nb_char) + 14;
          nb_char += length;
          break;
        case (ALPHANUMERIC_MODE):
          length = get_length(data,offset,mode);
          offset += decode_alphanumeric(data, offset, length, mes+nb_char) + 13;
          nb_char += length;
          break;
        case (END_MODE):
          end = 1;
          break;
        default:
          fprintf(stderr, "Cannot decode this mode (%d). Aborting...\n", mode);
          exit(-1);
      }
    }
    printf("%s\n", mes);

}


void test_etapes_correction(char *filename) {
  printf("fichier %s :\n", filename);
  QRCode qr = alloc_qrcode(SIZE_V1);
  pbm_to_qrcode_version1(filename, 12, 3, qr);
  int level, mask;
  get_format(qr, &level, &mask);
  data_mask(qr, mask);
  byte data[26];
  get_data_bytes(qr, data);

  int nb_data_bytes = NB_DATA_BYTES[level];
  int nb_correction_bytes = 26 - nb_data_bytes;

  /* syndromes des données */
  byte *S=malloc(nb_correction_bytes);
  data_syndrome(data, nb_correction_bytes, S);

  /* polynome localisateur d'erreurs */
  byte *L = malloc(1+nb_correction_bytes/2);
  int nb_errors = error_locator(nb_correction_bytes, S, L);
  printf("Il y a %d erreur(s)\n", nb_errors);
  for (int i=0; i<nb_errors; i++)
    printf("%02x, ", L[i]);
  printf("\n");
  printf("Le polynome de localisation est ");
  print_bytes_polynomial(L, nb_errors);

  /* recherche des zeros du polynome localisateur */
  byte *X = malloc(nb_errors);
  if (nb_errors != find_locations(nb_errors, L, X)) {
    fprintf(stderr,"Problem, couldn't find all errors... Aborting.\n");
    exit(-1);
  }
  printf("Les erreurs sont sur les coefficients de degré : ");
  for (int i=0; i<nb_errors; i++)
    printf("%d, ", X[i]);
  printf("\n");
  printf("Les erreurs sont sur les octets : ");
  for (int i=nb_errors; i>0; i--)
    printf("%d, ", 25-X[i-1]);
  printf("\n");

  /* recherche des valeurs des erreurs */
  byte *V=malloc(nb_errors);
  if (-1 == error_values(nb_correction_bytes, nb_errors, S, X, V)) {
    fprintf(stderr,"Couldn't solve for error values.. Aborting.\n");
    exit(-1);
  }
  printf("Les valeurs des erreurs sont : ");
  for (int i=0; i<nb_errors; i++)
    printf("%02x, ", V[i]);
  printf("\n");

  /* correction des erreurs */
  int i;
  for (int index=0; index<nb_errors; index++) {
    i = 25-X[index];
    printf(" - octet %2d : avant = %02x, ", i, data[i]);
    printf("correctif = %02x, valeur finale = %02x\n", V[index], data[i]^V[index]);
    data[i] ^= V[index];
  }
  printf("data final : ");
  for (int i=0; i<26; i++) {
    printf("%02x ", data[i]);
  }
  printf("\n");

  printf("\n");
}
void do_tests(void) {

  //create_mask_images();

  /*
  test_check_fixed("tests/exemple-ASCII.pbm");
  test_check_fixed("tests/exemple-alphanumerique.pbm");
  test_check_fixed("tests/exemple-numerique.pbm");
  */

  /*
  test_raw_format("tests/exemple-ASCII.pbm");
  test_raw_format("tests/exemple-alphanumerique.pbm");
  test_raw_format("tests/exemple-numerique.pbm");
  */

  /*
  test_format("tests/exemple-ASCII.pbm");
  test_format("tests/exemple-alphanumerique.pbm");
  test_format("tests/exemple-numerique.pbm");
  */

  /*
  test_level_mask("tests/exemple-ASCII.pbm");
  test_level_mask("tests/exemple-alphanumerique.pbm");
  test_level_mask("tests/exemple-numerique.pbm");
  */

  /*
  test_masking("tests/exemple-ASCII.pbm", "tests/exemple-ASCII-masked.pbm");
  test_masking("tests/exemple-alphanumerique.pbm", "tests/exemple-alphanumerique-masked.pbm");
  test_masking("tests/exemple-numerique.pbm", "tests/exemple-numerique-masked.pbm");
  */


  /*
  test_data("tests/exemple-ASCII.pbm");
  test_data("tests/exemple-alphanumerique.pbm");
  test_data("tests/exemple-numerique.pbm");
  */

  /*
  test_mode_length("tests/exemple-ASCII.pbm");
  test_mode_length("tests/exemple-alphanumerique.pbm");
  test_mode_length("tests/exemple-numerique.pbm");
  */


  /*
  test_solve_linear_system();
  */

  /*
  test_correct_data("tests/exemple-errors-OK.pbm");
  test_correct_data("tests/exemple-errors-1.pbm");  //1
  test_correct_data("tests/exemple-errors-2.pbm");  //1, 25
  test_correct_data("tests/exemple-errors-3.pbm");  //1, 17, 25
  test_correct_data("tests/exemple-errors-4.pbm");  //1, 17, 21, 25
  test_correct_data("tests/exemple-errors-5.pbm");  //1, 6, 17, 21, 25

  //test_correct_data("tests/exemple-test.pbm");
  */

  /*
  test_format("tests/exemple-TP2-1.pbm");
  test_format("tests/exemple-TP2-2.pbm");
  test_format("tests/exemple-TP2-3.pbm");
  */

  /*
  test_data_syndromes("tests/exemple-TP2-1.pbm");
  test_data_syndromes("tests/exemple-TP2-2.pbm");
  test_data_syndromes("tests/exemple-TP2-3.pbm");
  test_data_syndromes("tests/exemple-ASCII.pbm");
  */

  test_etapes_correction("tests/exemple-TP2-1.pbm");
  test_etapes_correction("tests/exemple-TP2-2.pbm");
  test_etapes_correction("tests/exemple-TP2-3.pbm");

}


/* vim: set tabstop=2 shiftwidth=2 : */
